#include <arpa/inet.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "util.h"
#include "ms.h"

//a memory stick is managed in 512-block areas of flash. Sony calls them segments
//first one has 0x1ee blocks used, all further ones have 0x1f0

#define BLOCKS_PER_SEGMENT					512
#define MAX_SEGMENTS						16
#define MAX_BAD_BLOCKS_PER_SEGMENT			15

#define BLOCKS_TO_SEARCH_FOR_BOOT_BLOCK		17

#define NUM_DATA_BLOCKS_FIRST_SEGMENT		0x1EE
#define NUM_DATA_BLOCKS_OTHER_SEGMENTS		0x1F0

#define MAX_DEVICE_BLOCKS					(NUM_DATA_BLOCKS_FIRST_SEGMENT + (MAX_SEGMENTS - 1) * NUM_DATA_BLOCKS_OTHER_SEGMENTS)

#define MS_OW_FLG_GOOD_BLOCK				0x80
#define MS_OW_FLG_DATA_READABLE				0x40		//cleared when data that should be here fails to be read
#define MS_OW_FLG_NEWEST					0x10

#define MS_MG_FLG_0x10						0x10
#define MS_MG_FLG_0x20						0x20
#define MS_MG_FLG_USEFUL					0x08		//cleared for blocks that need to be erased and reused
#define MS_MG_FLG_SYSTEM					0x04		//set for blocks that contain FTL's sector data

struct MsBootBlkNfo {
	uint32_t start;					//0x00	@ 0x170 0x17C 0x188 0x194
	uint32_t len;					//0x04	@ 0x174 0x180 0x18C 0x198
	uint8_t type;					//0x08	@ 0x178 0x184 0x190 0x19C
	uint8_t rfu[3];					//0x09	@ 0x179 0x185 0x191 0x19D
};									//0x0C

#define MS_BBI_TYPE_BAD_BLOCK_TAB			0x01
#define MS_BBI_TYPE_DEVICE_INFO				0x0A

#define MS_DEVICE_INFO_MAGIC	{0x01, 0x04, 0xdf, 0x4a, 0x01, 0xff, 0x1c, 0x04, 0x02, 0xd9, 0x01, 0xff, 0x18, 0x02, 0xdf, 0x01};

struct MsDeviceInfoBlock {					//info block of type MS_BBI_TYPE_DEVICE_INFO
	uint8_t magic[16];				//0x000
	uint8_t rfu0[0x0a];				//0x010
	char manufDeviceVer[27];		//0x01a	//three null-terminated strings
	uint8_t rfu1[0xcb];				//0x035
	
	
	//starting at 0x100, this basically follows the CF IDENTIFY COMMAND OUTPUT. see cfspc2_0_compact_flash.pdf
	uint16_t endianMarker;			//0x100	0x848a in whatever endianness the following data is
	uint16_t numCyl;				//0x102
	uint16_t unused_104;			//0x104
	uint16_t numHeads;				//0x106
	uint16_t unused_108[2];			//0x108
	uint16_t sectorsPerTrack;		//0x10c
	uint16_t sectorsPerCardHi;		//0x10e
	uint16_t sectorsPerCardLo;		//0x110
	uint16_t unused_112;			//0x112
	uint16_t serialNum[10];			//0x114	actual bytes. to become a string, write them BE!!!
	uint16_t unused_128[2];			//0x128
	uint16_t numEccBytes;			//0x12c
	uint16_t fwVer[4];				//0x12e	actual bytes. to become a string, write them BE!!!
	uint16_t modelName[20];			//0x136	actual bytes. to become a string, write them BE!!!
	
	//more here, see cfspc2_0_compact_flash.pdf
	
};	//0x200

struct MsBootBlock {				//BE
	uint16_t blkid;					//0x000 - 0x01 for boot block
	uint16_t ftlVer;				//0x002 - major.minor. must be >= 1.0, < 2.0, 
	uint8_t rfu0[0xB8];				//0x004
	uint8_t numBbis;				//0x0BC - how many entries in "bbi[]" are used
	uint8_t rfu1[0xB3];				//0x0BD
	struct MsBootBlkNfo bbi[4];		//0x170, only first two seem to ever be used, real number used is in .numBbis
									//[0].type must be 0x01
	
	uint8_t msClass;				//0x1A0 - memory sticks have 0x01 here
	uint8_t msSubclass;				//0x1A1 - memory sticks have 0x02 here
	uint16_t kbPerBlock;			//0x1A2 - only 8 and 16 are acceptable values
	uint16_t numBlocks;				//0x1A4 - powers of 2 from 512 to 8192 acceptable
	uint16_t usableBlocks;			//0x1A6 - multiples of 0x1f0 << n acceptable, for n = 0..4. this is the number of "sectors" we show to the user more or less
	uint16_t pageSize;				//0x1A8 - must be 512 for memorysticks
	uint8_t oobSize;				//0x1AA - must be 16 for memorysticks, but 6 ar eused for ECC so only 10 are avail to us
	uint8_t unk_1AB;				//0x1AB
	uint8_t manufTimeTzOfst;		//0x1AC
	uint8_t manufTimeYearHi;		//0x1AD
	uint8_t manufTimeYearLo;		//0x1AE
	uint8_t manufTimeMonth;			//0x1AF
	uint8_t manufDateDay;			//0x1B0
	uint8_t manufTimeHour;			//0x1B1
	uint8_t manufTimeMinute;		//0x1B2
	uint8_t manufTimeSecond;		//0x1B3
	uint32_t serialNum;				//0x1B4
	uint8_t assemblyMakerCode;		//0x1B8
	uint8_t assemblyModelCode[3];	//0x1B9
	uint16_t memManufCode;			//0x1BC
	uint16_t memModelCode;			//0x1BE
	uint32_t unk_1C0;				//0x1C0
	uint8_t vpp;					//0x1C4 (in decivolts)
	uint8_t vcc;					//0x1C5 (in decivolts)
	uint16_t controllerNum;			//0x1C6
	uint16_t controllerFunc;		//0x1C8 - set to 0x1001 for magic gate sticks
	uint8_t rfu2[0x09];				//0x1CA
	uint8_t interfaceType;			//0x1D3
	uint16_t controllerCode;		//0x1D4
	uint8_t formatType;				//0x1D6 - must be 0x01
	uint8_t unk_1D7;				//0x1D7
	uint8_t deviceType;				//0x1D8
	uint8_t rfu3[0x27];				//0x1D9
};

struct MsOob {
	uint8_t overwriFlag;			//0xf8 is set when valid data is written to a good block. set to 0x98 if we fail to read it while overwriting anothe rpag ein this block and need a copy of this one
	uint8_t manageFlag;				//0xff is set when valid data is written to a good block
	uint16_t logicalAddr;
	uint8_t rfu[FLASH_OOB_SIZE - 4];
};

struct BlockInfoTable {
	uint16_t *freeBlockListPtr;					//0x00		at runtime this points into freeBlocks[] where we expect to find the next free block. when we iterate, we check if it reaches freeBlocks[numFreeBlocks] and if so, cycle it back to freeBlocks[0]. when looking for  ablock we look at this pointer, advance as needed, when we find it leave pointing where we found, and replace the value with 0xffff. this means that on all searches exceot the fist we'll see 0xffff at the pointer
	uint32_t numFreeBlocks;						//0x04
	uint32_t numBadBlocks;						//0x08
	uint16_t freeBlocks[BLOCKS_PER_SEGMENT];	//0x0c
};


struct Ms {
	struct FlashChip *chip;
	uint16_t bootBlockNo;
	struct MsBootBlock bootBlock;
	uint8_t blockBuffer[32][FLASH_DATA_SIZE];
	bool wornOut;
	
	//table
	uint16_t blockMap[MAX_DEVICE_BLOCKS];
	struct BlockInfoTable blockInfo[MAX_SEGMENTS];
};

static uint32_t msPrvDeviceDataBlocks(const struct Ms* ms)
{
	return NUM_DATA_BLOCKS_FIRST_SEGMENT + NUM_DATA_BLOCKS_OTHER_SEGMENTS * (flashGetNumBlocks(ms->chip) / BLOCKS_PER_SEGMENT - 1);
}

uint32_t msGetSize(const struct Ms* ms)
{
	return msPrvDeviceDataBlocks(ms) * flashGetNumPagesPerBlock(ms->chip);
}

static struct Ms* msPrvMountWithBootBlock(struct FlashChip *chip, uint_fast16_t bootBlockNo)		//equivalent of MakeTable
{
	uint_fast16_t segment, numSegments = flashGetNumBlocks(chip) / BLOCKS_PER_SEGMENT, i, numBadBlockEntries, block, consideredBlockIdx = 0, blocksNow = NUM_DATA_BLOCKS_FIRST_SEGMENT, blocksSoFar = 0;
	uint16_t bbTab[FLASH_DATA_SIZE / sizeof(uint16_t)];
	struct Ms *ms = myAlloc(sizeof(*ms));
	struct BlockInfoTable *blkInfo;
	enum Readability rr;
	struct MsOob oob;
	
	ms->chip = chip;
	ms->bootBlockNo = bootBlockNo;
	
	//now that we have an allocated ms struct, read boot block into it so that we can refer to it later
	rr = flashReadPage(chip, bootBlockNo, 0, &ms->bootBlock, NULL);
	if (rr != Perfect && rr != Correctible)
		goto fail;
		
	//TODO: get CIS if we care (we do not)
	
	//get info on the bad block table and verify it. it must be the first entry and start at zero! It must be at most a page long!
	if (ms->bootBlock.bbi[0].type != MS_BBI_TYPE_BAD_BLOCK_TAB || ntohl(ms->bootBlock.bbi[0].start) != 0 || ntohl(ms->bootBlock.bbi[0].len) % sizeof(uint16_t) || ntohl(ms->bootBlock.bbi[0].len) > FLASH_DATA_SIZE)
		goto fail;
	numBadBlockEntries = ntohl(ms->bootBlock.bbi[0].len) / sizeof(uint16_t);
	fprintf(stderr, "%u factory bad blocks\n", (unsigned)numBadBlockEntries);
	
	//read the bad block table and verify its size matches what we expectfrom the boot block
	rr = flashReadPage(chip, bootBlockNo, 1, bbTab, NULL);
	if (rr != Perfect && rr != Correctible)
		goto fail;
	
	for (i = 0; i < sizeof(bbTab) / sizeof(*bbTab) && ntohs(bbTab[i]) != 0xffff; i++) {
		
		fprintf(stderr, "entry %u is 0x%04x\n", (unsigned)i, ntohs(bbTab[i]));
	}
	if (i != numBadBlockEntries) {
		fprintf(stderr, "bad block table does not match expected size!\n");
		ms->wornOut = true;
		numBadBlockEntries = 0;
	}
	
	//palm driver keeps blkMap and blkInfo in RAM and writes them out to the table in storage heap at the end of each iteration
	//we have on such needs so this logic is much simplified
	memset(ms->blockMap, 0xff, sizeof(ms->blockMap));
		
	for (segment = 0, blkInfo = ms->blockInfo; segment < numSegments; segment++, blocksSoFar += blocksNow, blocksNow = NUM_DATA_BLOCKS_OTHER_SEGMENTS, blkInfo++) {
		
		uint_fast16_t badBlocksInThisSegment = 0;
		uint_fast8_t curOverwriFlag;
		uint16_t *freeBlockListEnd = blkInfo->freeBlockListPtr = blkInfo->freeBlocks;
		
		for (block = 0; block < BLOCKS_PER_SEGMENT; block++, consideredBlockIdx++) {
			
			uint_fast16_t logicalAddr, existingMapping;
			uint16_t *existingMappingP;
			
			fprintf(stderr, " %2u.%3u (%5u)...", (unsigned)segment, (unsigned)block, (unsigned)consideredBlockIdx);
			
			//is it bad
			for (i = 0; i < numBadBlockEntries; i++) {
				
				if (ntohs(bbTab[i]) == consideredBlockIdx) { 
					badBlocksInThisSegment++;
					break;
				}
			}
			
			//if it is known bad, we have no more interest in it
			if (i != numBadBlockEntries) {
				fprintf(stderr, "factory bad\n");
				continue;
			}
			
			//not beig abe to read page0 even ignorring correction is also a sign of a bad block
			rr = flashReadPage(chip, consideredBlockIdx, 0, NULL, &oob);
			if (rr == Unreadable) {	//uncorrectible is accepted...
				
				fprintf(stderr, "unreadable bad\n");
				badBlocksInThisSegment++;
				continue;
			}
			
			//it might also be marked bad in the spare area of page 0
			curOverwriFlag = oob.overwriFlag;
			if (!(curOverwriFlag & MS_OW_FLG_GOOD_BLOCK)) {
				fprintf(stderr, "marked bad\n");
				badBlocksInThisSegment++;
				continue;
			}
			
			//must be set, else this is not a block under our management
			if (!(oob.manageFlag & MS_MG_FLG_SYSTEM)) {
				fprintf(stderr, "marked not data block\n");
				continue;
			}
			
			//needs retirement?
			if (!(oob.manageFlag & MS_MG_FLG_USEFUL)) {
				
				fprintf(stderr, "marked recycled\n");
				(void)flashEraseBlock(chip, consideredBlockIdx);
				badBlocksInThisSegment++;
				continue;
			}
			
			logicalAddr = ntohs(oob.logicalAddr);
			
			//free?
			if (logicalAddr == 0xffff) {
				
				fprintf(stderr, "free\n");
				*freeBlockListEnd++ = consideredBlockIdx;
				continue;
			}
			
			//is logical address invalid (out of range?) if so, erase and remember this block
			if (logicalAddr < blocksSoFar || logicalAddr - blocksSoFar >= blocksNow) {
				
				fprintf(stderr, "out of bounds\n");
				(void)flashEraseBlock(chip, consideredBlockIdx);
				*freeBlockListEnd++ = consideredBlockIdx;
				continue;
			}
			
			existingMappingP = ms->blockMap + logicalAddr;
			existingMapping = *existingMappingP;
			//first mapping we see? it is authoritative. we record it and move on
			if (existingMapping == 0xffff) {
				fprintf(stderr, "first seen mapping for logical block %u\n", (unsigned)logicalAddr);
				*existingMappingP = consideredBlockIdx;
				continue;
			}
			
			//not first mapping? we have more work to do
			if (!(curOverwriFlag & MS_OW_FLG_NEWEST) || flashReadPage(chip, existingMapping, 0, NULL, &oob) == Unreadable || (oob.overwriFlag & MS_OW_FLG_NEWEST)) {	//uncorrectible is accepted...
				
				fprintf(stderr, "better mapping for logical block %u\n", (unsigned)logicalAddr);
				
				(void)flashEraseBlock(chip, existingMapping);
				*freeBlockListEnd++ = existingMapping;
				*existingMappingP = consideredBlockIdx;
				
				///XXX: existing code didnt do this "continue" but it seems right
				continue;
			}
			
			fprintf(stderr, "worse mapping for logical block %u\n", (unsigned)logicalAddr);
			(void)flashEraseBlock(chip, consideredBlockIdx);
			*freeBlockListEnd++ = consideredBlockIdx;
		}
		
		blkInfo->numFreeBlocks = freeBlockListEnd - blkInfo->freeBlocks;
		blkInfo->numBadBlocks = badBlocksInThisSegment;
		fprintf(stderr, "this segment has %u free blocks and %u bad blocks\n", (unsigned)blkInfo->numFreeBlocks, (unsigned)blkInfo->numBadBlocks);
		
		/*
		//XXX: i have no idea why the stock driver does this or why it corrupts memory as it does it
		//not doing it works just as well, so let's not
		if (segment == numSegments - 1) {
			if (blkInfo->numFreeBlocks == 0) {
				
				fprintf(stderr, "original driver overwrite random memory here\n");
				goto fail;
			}
			blkInfo->freeBlockListPtr[-1] = 0xffff;
			blkInfo->numFreeBlocks--;
			blkInfo->numBadBlocks++;
		}
		*/
		if (blkInfo->numBadBlocks > MAX_BAD_BLOCKS_PER_SEGMENT)
			ms->wornOut = true;
		
	}
	
	return ms;

fail:
	free(ms);
	return NULL;
}

struct Ms* msMount(struct FlashChip *chip)
{
	struct MsBootBlock data;
	enum Readability rr;
	uint_fast16_t block;
	struct MsOob oob;
	struct Ms *ms;
	
	//find a readable boot block
	for (block = 0; block < BLOCKS_TO_SEARCH_FOR_BOOT_BLOCK; block++) {
		
		uint_fast16_t val16;
		
		rr = flashReadPage(chip, block, 0, &data, &oob);
		
		fprintf(stderr, "Read %u.0 -> %u\n", (unsigned)block, rr);
		
		//must be readable
		if (rr != Perfect && rr != Correctible)
			continue;
		
		//cannot be set
		if (oob.manageFlag & MS_MG_FLG_SYSTEM)
			continue;
		
		//must be 0x0001
		if (ntohs(data.blkid) != 0x0001)
			continue;
		
		//FTL version must be one we know about
		if (ntohs(data.ftlVer) < 0x0100 || ntohs(data.ftlVer) >= 0x0200)
			continue;
		
		//class & subclass must be correct
		if (data.msClass != 0x01 || data.msSubclass != 0x02)
			continue;
		
		//block size must be one MS FTL supports
		if (ntohs(data.kbPerBlock) != 8 && ntohs(data.kbPerBlock) != 16)
			continue;
		
		//num blocks must be a power of two between 512 and 8192
		val16 = ntohs(data.numBlocks);
		if (val16 < 512 || val16 > 8192 || (val16 & (val16 - 1)))
			continue;
		
		//num usable blocks must be a power of two multiple of 0x1f0 between 0x1f0 and 0x1f00
		//multiple must match that of numBlocks
		val16 = ntohs(data.usableBlocks);
		if (val16 < 0x1f0 || val16 > 0x1f00 || val16 % 0x1f0 || val16 / 0x1f0 != ntohs(data.numBlocks) / 512)
			continue;
		
		//page size and OOB size must be what MS requires
		if (ntohs(data.pageSize) != 512 || data.oobSize != 16)
			continue;
		
		//format type must be proper for ms
		if (data.formatType != 0x01)
			continue;
		
		//one of the address records' type is also verified
		if (data.bbi[0].type != MS_BBI_TYPE_BAD_BLOCK_TAB)
			continue;
		
		fprintf(stderr, "Found a boot block at index %lu. An %s with %u blocks (%u usable) of %uKB each made on "
				"%04u/%02u/%02u at %02u:%02u:%02u. Snum: %lu\n",
				(unsigned long)block, ntohs(data.controllerFunc) == 0x1001 ? "MG" : "MS", ntohs(data.numBlocks),
				ntohs(data.usableBlocks), ntohs(data.kbPerBlock), (unsigned)((unsigned)data.manufTimeYearHi * 256 + data.manufTimeYearLo),
				data.manufTimeMonth, data.manufDateDay, data.manufTimeHour, data.manufTimeMinute, data.manufTimeSecond,
				(unsigned long)ntohl(data.serialNum));
		
		if (ntohs(data.numBlocks) != flashGetNumBlocks(chip) || ntohs(data.kbPerBlock) * (1024 / FLASH_DATA_SIZE) != flashGetNumPagesPerBlock(chip)) {
			fprintf(stderr, "This does not match flash geometry. Ignoring!\n");
			continue;
		}
		
		return msPrvMountWithBootBlock(chip, block);
	}
	
	fprintf(stderr, "Failed to find a valid boot block\n");
	
	return NULL;
}

bool msRead(struct Ms *ms, uint32_t lba, void *data)
{
	uint32_t virtBlockIdx = lba / flashGetNumPagesPerBlock(ms->chip);
	uint8_t pageIdx = lba % flashGetNumPagesPerBlock(ms->chip);
	uint_fast16_t physBlockIdx;
	enum Readability rr;
	struct MsOob oob;
	
	if (virtBlockIdx >= msPrvDeviceDataBlocks(ms))
		return false;
	
	physBlockIdx = ms->blockMap[virtBlockIdx];
	if (physBlockIdx == 0xffff) {
		
		memset(data, 0xff, MS_SECTOR_SIZE);
		return true;
	}
	
	rr = flashReadPage(ms->chip, physBlockIdx, pageIdx, data, &oob);
	if (rr != Perfect && rr != Correctible)
		return false;
	
	//OPTIONAL: mark the error so we dont keep hitting it. 
	//this seems to be pointless since that just causes the same kind of error
	//to be returned. so why bother?
	
	return (oob.manageFlag & (MS_MG_FLG_0x10 | MS_MG_FLG_0x20)) == (MS_MG_FLG_0x10 | MS_MG_FLG_0x20);
}

void msUnmount(struct Ms *ms)
{
	free(ms);
}


